/* usadelndsoc
 *
 * Copyright © 2022 Pauli Virtanen
 *    @author Pauli Virtanen <pauli.t.virtanen@jyu.fi>
 *
 * SPDX-License-Identifier: AGPL-3.0-or-later
 */
#include <pybind11/pybind11.h>
#include <pybind11/complex.h>

#include "common.hpp"
#include "array.hpp"
#include "action.hpp"

#include "pythonutil.hpp"

using namespace array;
using namespace array::python;

#define CLASS_Action PY_VISIBILITY Action
class CLASS_Action
{
public:
    typedef PyArrayView<const Complex,Shape<Dynamic,Dynamic,4,4,4> > UType;
    typedef PyArrayView<const Complex,Shape<Dynamic,Dynamic,4,4> > OmegaType;
    typedef PyArrayView<const Complex,Shape<Dynamic,Dynamic,2,2,2> > QType;
    typedef PyArrayView<const Mask,Shape<Dynamic,Dynamic> > MaskType;

private:
    UType U_;
    OmegaType Omega_;
    MaskType mask_;
    size_t nx_, ny_;
    double Lx_, Ly_;
    Complex alpha_;
    Complex omega_;

    typedef Vector<size_t> SizeVector;
    typedef Vector<Complex> ComplexVector;
    typedef std::vector<bool> BoolVector;
    typedef CppAD::sparse_rc<SizeVector> Sparsity;

    CppAD::ADFun<Complex> f_;
    CppAD::sparse_hes_work hes_work_;
    Sparsity hes_pattern_;
    CppAD::sparse_rcv<SizeVector, ComplexVector> hes_subset_;
    bool computed_ = false;

    pybind11::object py_sparse_matrix_type_;

    static Vector<Complex> ravel(const QType& Q)
        {
            Complex *data = const_cast<Complex *>(Q.data());
            size_t size = Q.size();
            return std::move(Vector<Complex>(data, size));
        }

    void set_dynamic()
        {
            std::vector<Complex> dynamic(1);
            dynamic[0] = omega_;
            f_.new_dynamic(dynamic);
        }

public:
    Action(double Lx, double Ly, Complex alpha, MaskType&& mask, UType&& U, OmegaType&& Omega)
        : U_(std::move(U)), Omega_(std::move(Omega)), mask_(std::move(mask)), nx_(U_.dim(0)), ny_(U_.dim(1)), Lx_(Lx), Ly_(Ly), alpha_(alpha), omega_(0)
        {
            if (Omega.dim(0) != nx_ || Omega.dim(1) != ny_)
                throw std::domain_error("U and Omega have incompatible shape");
            if (mask.dim(0) != nx_ || mask.dim(1) != ny_)
                throw std::domain_error("U and mask have incompatible shape");

            pybind11::object sparse_module = pybind11::module_::import("scipy.sparse");
            py_sparse_matrix_type_ = sparse_module.attr("coo_matrix");
        }

    void set(const Complex& omega)
        {
            omega_ = omega;
        }

    Complex eval(const QType& Q)
        {
            if (Q.dim(0) != nx_ || Q.dim(1) != ny_)
                throw std::out_of_range("Q array wrong size");
            return S(mask_, Q, omega_, U_, Omega_, Lx_, Ly_, alpha_);
        }

    CppAD::ADFun<Complex> eval_ad(const QType& Qval)
        {
            if (Qval.dim(0) != nx_ || Qval.dim(1) != ny_)
                throw std::out_of_range("Q array wrong size");

            Array<ADComplex, Shape<Dynamic,Dynamic,2,2,2> >
                Q({nx_, ny_, 2, 2, 2});

            auto a = Qval.reshape<Dynamic>({Qval.size()});
            auto b = Q.reshape<Dynamic>({Q.size()});
            for (size_t i = 0; i < a.dim(0); ++i)
                b(i) = a(i);

            std::vector<ADComplex> dynamic(1);
            dynamic[0] = omega_;

            CppAD::Independent(Q.storage(), dynamic);

            Array<ADComplex, Shape<1> > res;
            res(0u) = S(mask_, Q, dynamic[0], U_, Omega_, Lx_, Ly_, alpha_);

            return {Q.storage(), res.storage()};
        }

    void compute(const QType& Q)
        {
            f_ = eval_ad(Q);
            f_.optimize();

            /*
             * Hessian sparsity pattern
             */

            hes_pattern_.resize(0, 0, 0);
            hes_work_.clear();

            /* Right-multiplier is identity matrix */
            size_t n      = Q.size();
            size_t nr     = n;
            size_t nc     = n;
            size_t nnz_in = n;
            Sparsity pattern_in(nr, nc, nnz_in);
            for(size_t k = 0; k < nnz_in; k++)
            {
                size_t r = k;
                size_t c = k;
                pattern_in.set(k, r, c);
            }

            /* Hessian sparsity: */
            BoolVector s = {true};
            f_.for_jac_sparsity(pattern_in, false, false, false, hes_pattern_);
            f_.rev_hes_sparsity(s, false, false, hes_pattern_);
            f_.size_forward_set(0);

            hes_subset_ = hes_pattern_;

            computed_ = true;
        }

    auto grad(const QType& Q)
        {
            Vector<Complex> x = ravel(Q);

            if (!computed_)
                compute(Q);
            else
                set_dynamic();

            auto jac = f_.Jacobian(x);
            return Array<Complex, Shape<Dynamic>, Vector<Complex> >(std::move(jac), {jac.size()});
        }

    auto hess(const QType& Q)
        {
            Vector<Complex> x = ravel(Q);

            if (!computed_)
                compute(Q);
            else
                set_dynamic();

            /* Hessian */
            ComplexVector w = {1.0};
            std::string coloring = "cppad.symmetric";
            f_.sparse_hes(x, w, hes_subset_, hes_pattern_, coloring, hes_work_);

            size_t nnz = hes_subset_.nnz();

            Array<size_t, Shape<Dynamic>, SizeVector> row(std::move(SizeVector(hes_subset_.row())), {nnz});
            Array<size_t, Shape<Dynamic>, SizeVector> col(std::move(SizeVector(hes_subset_.col())), {nnz});
            Array<Complex, Shape<Dynamic>, ComplexVector> val(std::move(ComplexVector(hes_subset_.val())), {nnz});

            return py_sparse_matrix_type_(
                pybind11::make_tuple(std::move(val), pybind11::make_tuple(std::move(row), std::move(col))),
                pybind11::make_tuple(x.size(), x.size()));
        }
};

PYBIND11_MODULE(_core, m)
{
    namespace py = pybind11;

    import_array1();

    m.doc() = "usadelndsoc._core";

    py::class_<Action>(m, "Action")
        .def(py::init<double, double, Complex, Action::MaskType, Action::UType, Action::OmegaType>(), "Action",
             py::arg("L_x"), py::arg("L_y"), py::arg("alpha"), py::arg("mask"), py::arg("U"), py::arg("Omega"))
        .def("set", &Action::set, "Set parameters", py::arg("omega"))
        .def("eval", &Action::eval, "Evaluate value of action", py::arg("Q"))
        .def("grad", &Action::grad, "Evaluate gradient of action", py::arg("Q"))
        .def("hess", &Action::hess, "Evaluate hessian of action", py::arg("Q"))
        .def("compute", &Action::compute, "Precompute for AD", py::arg("Q"))
        ;

    m.attr("MASK_NONE") = static_cast<int>(Mask::NONE);
    m.attr("MASK_TERMINAL") = static_cast<int>(Mask::TERMINAL);
    m.attr("MASK_VACUUM") = static_cast<int>(Mask::VACUUM);
}
